#include <cassert>
#include <stdio.h>

#include <windows.h>

#include "idx.h"

// "is letter" bitfield. includes umlaut and  (ISO-8859-1)
// LSB 5 bits select bit position; next 3 index into table.
// extra work (2x shift, 2x and) is worth it: table fits in one cache line.
//
// could use std::bitset, but initialization is clunky
static const u32 is_alpha_tbl[256/32] =
{
	0x00000000,	0x00000000,
	0x07fffffe,	0x07fffffe,
	0x00000000,	0x00000000,
	0x90400010,	0x10400010
};

static inline bool _is_alpha(const u8 c)
{
	return (is_alpha_tbl[c >> 5] & (1ul << (c & 0x1f))) != 0;
}








float score;
float bias = 0;

float iir(float f, float sharpness = 0.7f)
{
	return bias = f * sharpness + bias * (1.0f - sharpness);
}

bool accept_word(const char* word, int len)
{
	if(len < MIN_WORD_LEN)
		return false;
	assert(len > 0);

	static bool freqs_loaded;

	// first call - read freqs from file
	static u8* trigram_freqs;
	if(!trigram_freqs)
	{
		trigram_freqs = (u8*)VirtualAlloc(0, 32768, MEM_COMMIT, PAGE_READWRITE);

		// read module may have changed dir - make sure we're reading from startup dir
		static char path[MAX_PATH+1];
		strcpy(path, __argv[0]);
		strcpy(strrchr(path, '\\')+1, "trigram_freqs.dat");

		HANDLE hFile = CreateFile(path, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
		if(hFile != INVALID_HANDLE_VALUE)
		{
			u32 bytes_read;
			ReadFile(hFile, trigram_freqs, 32768, &bytes_read, 0);
			CloseHandle(hFile);

			freqs_loaded = true;
		}
	}

	if(!freqs_loaded)
		return true;

	score = 2.0f;

	unsigned idx = 0;
	for(int i = 0; i < len; i++)
	{
		idx &= 0x3ff;	// clear upper 5 bits
		idx <<= 5;
		unsigned next = word[i]-'A';
		if(next > 26)
		{
			if(next == (unsigned)(''-'A')) next = 26; else
			if(next == (unsigned)(''-'A')) next = 27; else
			if(next == (unsigned)(''-'A')) next = 28; else
			if(next == (unsigned)(''-'A')) next = 29;
		}
		idx |= next;

		if(i >= 2)
		{
			u8 freq = trigram_freqs[idx];

			if(!freq)
				score += -60.0f;
			else if(freq <= 2)
				score += -7.0f;
			else
				score += freq;
		}
	}

	score /= len;

	// decent word
	if(score > 15.0f)
	{
		// get out of the dumps
		if(bias < -10.0f)
			bias = -10.0f;

		// pretty good word
		if(score > 25.0f)
			bias += 4.0f;

		bias += 3.0f;
	}
	// not good
	else
	{
		// in doubt
		if(bias > 3.0f)
			bias = 3.0f;

		// fairly bad word
		if(score < 8.0)
			bias -= 4.0f;

		bias -= 2.0f;
	}

	return (score + bias) > 13.0f;

/*
	// decent word
	if(score > 10.0f)
	{
		if(bias < -7.0f)
			bias = -7.0f;
	}
	else
	{
		if(bias > -3.0f)
			bias = -3.0f;
	}

	if(score+bias > 13.0f)
	{
		if(bias < 20.0f)
			bias += 2.0f;
		return true;
	}

	bias -= 6.0f;
	return false;
*/
}



extern u32 hash(const char* str, int len);
#include <set>



long extract_words(const u8* buf, u32 size)
{
bias=0;

#ifdef DUMP_AR
static FILE* fr;
if(!fr)
{
fr=fopen("d:\\projects\\info4\\idx\\reject.csv", "w");
setvbuf(fr, 0, _IONBF, 0);
}
#endif


	static char word[MAX_WORD_LEN+1];
	static int len;

	static int h;// allow start - \n end hyphenation
	// 1: last char was hyphen; 2: hyphen and newline(s)

	int last_word_end=0;
	int pos = 0;	// in buf



	// caller is telling us EOF reached - finish off partial word
	if(!buf)
	{
		lookup_word(word, len);
		len = 0;
		return 0;
	}

	bias = 0;

	// for each char in buffer:
	while(size--)
	{
		const unsigned int c = buf[pos++];
		const bool is_alpha = _is_alpha(c);

		if(len)
		{
			// possible hyphen
			if(c == '-')
				h = 1;
			// hyphen + newline
			else if(h > 0 && (c == '\n' || c == '\r'))
				h = 2;
			// word terminator
			else if(!is_alpha || h == 1 || len >= MAX_WORD_LEN)
			{
/*
score for each word:
 space = good, caps in text = bad
    xngram

state (quality of block)
 # whitespace / text
    xhow long since last accepted word
*/


				bool done = false;
				word[len]= 0 ;


				// more than 20 chars since last accepted word
/*				int dist = pos - last_word_end;
				if(dist > 20)
					bias = -10.0f;
				else if(dist < 10)
					bias = +6.0f;
*/				
if(len >= MIN_WORD_LEN)
				if(accept_word(word, len))
				{
					last_word_end = pos;
					if(lookup_word(word, len) == -1)	// telling us to stop
						done = true;
				}
#ifdef DUMP_AR
else
{
u32 h = hash(word, len);
static std::set<u32> rejs;
if(rejs.find(h) == rejs.end())
{
rejs.insert(h);
fprintf(fr, "%.2f, %.2f, %s\n", score, bias, word);
}
}
#endif
				h = 0;
				len = 0;
				if(done)
					return last_word_end;
			}
		}

		if(is_alpha)
			word[len++] = c & ~0x20;// uppercase (=> no special case for '' == 0xdf)
	}

	return last_word_end;
}


static long mp3_scan(const u8* buf, u32 size)
{
	assert(buf != 0);
	

	if (size >= 128)
		buf += size-128;
	else
		return extract_words(buf, size);
		

	int i = 0;
	if (((*buf) == 'T') && ((*(buf+1) == 'A')) && (*(buf+2) == 'G') && (size >= 128)) // it is the TAG begin...
	{
		// get "Song Title" - Tag [30 Bytes - leftover space is filled with binary 0] and extractwords
		buf += 3;
		while ((buf + i) && (i < 30) )  {
            i++;
		}
		extract_words(buf, i-1);

		// get "Artist" - Tag [30 Bytes - leftover space is filled with binary 0] and extractwords
		i = 0;
		buf += 30;
		while ((buf + i) && (i < 30) )  {
            i++;
		}
		extract_words(buf, i-1);

		// get "Album" - Tag [30 Bytes - leftover space is filled with binary 0] and extractwords - skip the year TAG [4 Bytes]
		i = 0;
		buf += 30;
		while ((buf + i) && (i < 30) )  {
            i++;
		}
		extract_words(buf, i-1);

		// get "Comment" - Tag [30 Bytes - leftover space is filled with binary 0] and extractwords 
		i = 0;
		buf += 34;
		while ((buf + i) && (i < 30) )  {
            i++;
		}
		extract_words(buf, i-1);

		// the last byte contains information about genre of this song. we leave this for professional mp3 players
	}

        return 0;
}


long htmlpspdf_scan(const u8* buf, u32 size, bool is_pspdf)
{
int pos = 0;
int last_word_end = 0;
     static int is_in_tag;
     static int len;
     static int h;
     static char word[MAX_WORD_LEN+1];
     static int forbidden_char = 0;
     int is_alpha = 0;
	 static bool maybe_corrected = false;
	 static bool mustbe_corrected = false;

     // caller is telling us EOF reached - finish word in buffer
	 if(!buf)
	 {
          lookup_word(word, len);
          len = 0;
		  return -1;
     }

    while (size--) {
        unsigned int t = buf[pos++];
        is_alpha = _is_alpha(t);

		// is it the beginning of a non breaking space or an Umlaut
		if ((t == '&') && (!is_pspdf)) {
			forbidden_char = 1;
			is_alpha = 1;
			continue;
		}

		// start of a tag => look up buffer
        if (t == '<') {
                lookup_word(word, len);
                len = 0;
                is_in_tag = 1;
        }

		// leave tag, allow letters to be written again into the buffer
        if (t == '>') is_in_tag = 0;



		// check for non breaking spaces and Umlaut and correct buffer if necessary
        if ((t == ';') && forbidden_char && !is_pspdf) {
                if (len - 4 >= 0) {

			  		// is it a non breaking space [&nbsp;]
                    if ((word[len-4] == 'N') && (word[len-3] == 'B') &&
						(word[len-2] == 'S') && (word[len-1] == 'P')) { len -= 4;
						} else
					// is it a quotation [&quot;]
					if ((word[len-4] == 'Q') && (word[len-3] == 'U') &&
						(word[len-2] == 'O') && (word[len-1] == 'T')) { 
							len -= 4;
						} else 
					// is it a "" or "" [&(a/A)uml;] - if yes enter correct Umlaut
					if ((word[len-4] == 'A') && (word[len-3] == 'U') &&
						(word[len-2] == 'M') && (word[len-1] == 'L')) {
                            t = (unsigned)'';
							len -= 4;
							is_alpha = 1;
						} else 

					// is it a "" or "" [&(o/O)uml;] - if yes enter correct Umlaut
					if ((word[len-4] == 'O') && (word[len-3] == 'U') &&
						(word[len-2] == 'M') && (word[len-1] == 'L')) {
                            t = (unsigned)'';
							len -= 4;
							is_alpha = 1;
						} else

					// is it a "" or "" [&(u/U)uml;] - if yes enter correct Umlaut
					if ((word[len-4] == 'U') && (word[len-3] == 'U') &&
						(word[len-2] == 'M') && (word[len-1] == 'L')) {
                            t = (unsigned)'';
							len -= 4;
							is_alpha = 1;
						}
				}

				if (len - 5 >= 0 ) {
					// is it a "" [&szlig;] - if yes exchange encoded Umlaut by 
					if ((word[len-5] == 'S') && (word[len-4] == 'Z') &&
						(word[len-3] == 'L') && (word[len-2] == 'I') && (word[len-1] == 'G')) {
                            t = (unsigned)'';
							len -= 5;
							is_alpha = 1;
						}
                }
                forbidden_char = 0;
        }
/*
		if (is_pspdf) {
			// in case it was a not an Umlaut
			if (mustbe_corrected && ((t != 'a') && (t != 'A') && (t != 'o') && (t != 'O') && (t != 'u') && (t != 'U'))) {
				len -= 2; 
				mustbe_corrected = false;
				maybe_corrected = false;						
			}						
			
			// beginning character fpr ps / pdf messed up Umlaut
			if ((t == 0x7f)|| (t == 0xffffff98))  {
				maybe_corrected = true; 
				is_alpha = 1;
			}

			// second character
			if ((t ==' ') && maybe_corrected) { 
				mustbe_corrected = true;
				maybe_corrected = false;
				is_alpha = 1;
			}

			// correct if is an 
			if (mustbe_corrected && ((t == 'a') || (t == 'A'))) {
				mustbe_corrected = false;
				maybe_corrected = false;
				t = '';
				len -= 2;
			}
				
			// correct if it is an 
			if (mustbe_corrected && ((t == 'o') || (t == 'O'))) {
				mustbe_corrected = false;
				maybe_corrected = false;
				t = '';
				len -= 2;
			}

			// ...PS way for 
			if (maybe_corrected && ((t == 'o') || (t == 'O'))) {
				mustbe_corrected = false;
				maybe_corrected = false;
				t = '';
				len -= 1;
			}

			// correct if it is an 
			if (mustbe_corrected && ((t == 'u') || (t == 'U'))) {
				mustbe_corrected = false;
				maybe_corrected = false;
				t = '';
				len -= 2;
			}

			// ...PS way for 
			if (t == 0x19) {
				t = ''; 
				is_alpha = 1;
			}

			// if the second character is a newline instead of a space
			if ((t =='\n') && maybe_corrected) { 
				mustbe_corrected = true;
				maybe_corrected = false;
				is_alpha = 1;
			}

  		}
*/


        if (len && !is_in_tag)
		{
            // lensible hyphen
            if(t == '-')
                h = 1;
            // hyphen + newline
            else if(h > 0 && (t == '\n' || t == '\r'))
                h = 2;
            // 0 byte in UTF-16
            else if(t == 0)
                ;
            // word terminator
            else if((!is_alpha && !maybe_corrected && !mustbe_corrected) || h == 1 || len >= MAX_WORD_LEN)
            {
			word[len]=0;
				bool done = false;
                if(lookup_word(word, len) == -1)
					done = true;
                last_word_end = pos;
                len = 0;
                h = 0;
			if(done)
				return last_word_end;
            }
        }

		// if it is a letter and it is not within a tag or if it is in ps / pdf add letter to buffer 
        if ((is_alpha && !is_in_tag) || (is_pspdf && is_alpha))
                    word[len++] = t & ~0x20;     // uppercase (=> no special case for '' == 0xdf)

    }

	
    return last_word_end;
}



long process_buf(const u8* buf, u32 size, FileType type)
{
	switch(type)
	{
	case FT_PDF:
	case FT_HTML:
		return htmlpspdf_scan(buf, size, type == FT_PDF);

	case FT_MP3:
		return mp3_scan(buf, size);

	default:
		return extract_words(buf, size);
	}
}
